'''An asynchronous multi-process `HTTP proxy server`_. It works for both
``http`` and ``https`` (tunneled) requests.


Managing Headers
=====================
It is possible to add middleware to manipulate the original request headers.
If the header middleware is
an empty list, the proxy passes requests and responses unmodified.
This is an implementation for a forward-proxy which can be used
to retrieve any type of source from the Internet.

To run the server::

    python manage.py

An header middleware is a callable which receives the wsgi *environ* and
the list of request *headers*. By default the example uses:

.. autofunction:: x_forwarded_for

To run with different headers middleware create a new script and do::

    from proxyserver.manage import server

    if __name__ == '__main__':
        server(headers_middleware=[...]).start()

Implemenation
===========================

.. autoclass:: ProxyServerWsgiHandler
   :members:
   :member-order:


.. _`HTTP proxy server`: http://en.wikipedia.org/wiki/Proxy_server
'''
import logging
from functools import partial
import asyncio

import pulsar
from pulsar import HttpException, task
from pulsar.apps import wsgi, http
from pulsar.apps.http.plugins import noerror
from pulsar.utils.httpurl import Headers, ENCODE_BODY_METHODS
from pulsar.utils.log import LocalMixin, local_property


SERVER_SOFTWARE = 'Pulsar-proxy-server/%s' % pulsar.version
ENVIRON_HEADERS = ('content-type', 'content-length')
USER_AGENT = SERVER_SOFTWARE
logger = logging.getLogger('pulsar.proxyserver')


def x_forwarded_for(environ, headers):
    '''Add *x-forwarded-for* header'''
    headers.add_header('x-forwarded-for', environ['REMOTE_ADDR'])


class ProxyServerWsgiHandler(LocalMixin):
    '''WSGI middleware for an asynchronous proxy server.

    To perform processing on headers you can pass a list of
    ``headers_middleware``.
    An headers middleware is a callable which accepts two parameters, the wsgi
    *environ* dictionary and the *headers* container.
    '''
    def __init__(self, headers_middleware=None):
        self.headers_middleware = headers_middleware or []

    @local_property
    def http_client(self):
        '''The :class:`.HttpClient` used by this proxy middleware for
        accessing upstream resources'''
        client = http.HttpClient(decompress=False, store_cookies=False)
        client.headers.clear()
        return client

    def __call__(self, environ, start_response):
        uri = environ['RAW_URI']
        logger.debug('new request for %r' % uri)
        if not uri or uri.startswith('/'):  # No proper uri, raise 404
            raise HttpException(status=404)
        response = TunnelResponse(self, environ, start_response)
        response.request()
        return response.future


############################################################################
#    RESPONSE OBJECTS
class TunnelResponse:
    '''Base WSGI Response Iterator for the Proxy server
    '''
    def __init__(self, wsgi, environ, start_response):
        self.wsgi = wsgi
        self.environ = environ
        self.start_response = start_response
        self.future = asyncio.Future()

    @task
    def request(self):
        '''Perform the Http request to the upstream server
        '''
        request_headers = self.request_headers()
        environ = self.environ
        method = environ['REQUEST_METHOD']
        data = None
        if method in ENCODE_BODY_METHODS:
            data = DataIterator(self)
        http = self.wsgi.http_client
        try:
            yield from http.request(method,
                                    environ['RAW_URI'],
                                    data=data,
                                    headers=request_headers,
                                    version=environ['SERVER_PROTOCOL'],
                                    pre_request=self.pre_request)
        except Exception as exc:
            self.error(exc)

    def request_headers(self):
        '''Fill request headers from the environ dictionary and
        modify them via the list of :attr:`headers_middleware`.
        The returned headers will be sent to the target uri.
        '''
        headers = Headers(kind='client')
        for k in self.environ:
            if k.startswith('HTTP_'):
                head = k[5:].replace('_', '-')
                headers[head] = self.environ[k]
        for head in ENVIRON_HEADERS:
            k = head.replace('-', '_').upper()
            v = self.environ.get(k)
            if v:
                headers[head] = v
        for middleware in self.wsgi.headers_middleware:
            middleware(self.environ, headers)
        return headers

    def error(self, exc):
        if self.future.done():
            self.future.set_exception(exc)
        else:
            logger.error(str(exc))

    @noerror
    def pre_request(self, response, exc=None):
        '''Start the tunnel.

        This is a callback fired once a connection with upstream server is
        established.
        '''
        if response.request.method == 'CONNECT':
            # proxy - server connection
            upstream = response.connection
            # client - proxy connection
            dostream = self.environ['pulsar.connection']
            # Upgrade downstream connection
            dostream.upgrade(partial(StreamTunnel, upstream))
            # Upgrade upstream connection
            upstream.upgrade(partial(StreamTunnel, dostream))
            self.start_response('200 Connection established', [])
            # send empty byte so that headers are sent
            self.future.set_result([b''])
            response.abort_request()
        else:
            response.bind_event('data_processed', self.data_processed)
            response.bind_event('post_request', self.post_request)

    def data_processed(self, response, data=None, **kw):
        self.environ['pulsar.connection'].write(data)

    def post_request(self, _, exc=None):
        self.future.set_exception(wsgi.AbortWsgi())


class DataIterator:

    def __init__(self, response):
        self.response = response
        self.stream = response.environ.get('wsgi.input')

    def __iter__(self):
        yield self.stream.reader.read()


class StreamTunnel(pulsar.ProtocolConsumer):
    ''':class:`.ProtocolConsumer` handling encrypted messages from
    downstream client and upstream server.

    This consumer is created as an upgrade of the standard Http protocol
    consumer.

    .. attribute:: tunnel

        Connection to the downstream client or upstream server.
    '''
    headers = None
    status_code = None
    http_request = None

    def __init__(self, tunnel, loop=None):
        super().__init__(loop)
        self.tunnel = tunnel

    def connection_made(self, connection):
        self.logger.debug('Tunnel connection %s made', connection)
        connection.bind_event('connection_lost', self._close_tunnel)
        if self.http_request:
            self.start(self.http_request)

    def data_received(self, data):
        try:
            return self.tunnel.write(data)
        except Exception:
            if not self.tunnel.closed:
                raise

    def _close_tunnel(self, arg, exc=None):
        if not self.tunnel.closed:
            self._loop.call_soon(self.tunnel.close)


def server(name='proxy-server', headers_middleware=None,
           server_software=None, **kwargs):
    '''Function to Create a WSGI Proxy Server.'''
    if headers_middleware is None:
        headers_middleware = [x_forwarded_for]
    wsgi_proxy = ProxyServerWsgiHandler(headers_middleware)
    kwargs['server_software'] = server_software or SERVER_SOFTWARE
    return wsgi.WSGIServer(wsgi_proxy, name=name, **kwargs)


if __name__ == '__main__':
    server().start()



'''Tests the tools and utilities in pulsar.utils.'''
import unittest
from datetime import datetime, date

from pulsar import system, get_actor
from pulsar.utils.tools import checkarity, nice_number, date2timestamp
from pulsar.utils.importer import py_file, import_modules
from pulsar.apps.test import sequential


def f0(a, b):
    pass


def f0_discount(request, a, b):
    pass


def f1(a, b=0):
    pass


def f2(a, **kwargs):
    # This fails curretly
    pass


def arity_check(func, *args, **kwargs):
    discount = kwargs.pop('discount', 0)
    return checkarity(func, args, kwargs, discount=discount)


class TestArityCheck(unittest.TestCase):

    def testArity0(self):
        self.assertEqual(arity_check(f0, 3, 4), None)
        self.assertEqual(arity_check(f0, 3),
                         '"f0" takes 2 parameters. 1 given.')
        self.assertEqual(arity_check(f0),
                         '"f0" takes 2 parameters. 0 given.')
        self.assertEqual(arity_check(f0, 4, 5, 6),
                         '"f0" takes 2 parameters. 3 given.')
        self.assertEqual(arity_check(f0, a=3, b=5), None)
        self.assertEqual(arity_check(f0, a=3, c=5),
                         '"f0" has missing "b" parameter.')
        self.assertEqual(arity_check(f0, a=3, c=5, d=6),
                         '"f0" takes 2 parameters. 3 given.')

    def testArity0WidthDiscount(self):
        f0 = f0_discount
        fname = f0.__name__
        self.assertEqual(arity_check(f0, 3, 4, discount=1), None)
        self.assertEqual(arity_check(f0, 3, discount=1),
                         '"%s" takes 2 parameters. 1 given.' % fname)
        self.assertEqual(arity_check(f0, discount=1),
                         '"%s" takes 2 parameters. 0 given.' % fname)
        self.assertEqual(arity_check(f0, 4, 5, 6, discount=1),
                         '"%s" takes 2 parameters. 3 given.' % fname)
        self.assertEqual(arity_check(f0, a=3, b=5, discount=1), None)
        self.assertEqual(arity_check(f0, a=3, c=5, discount=1),
                         '"%s" has missing "b" parameter.' % fname)
        self.assertEqual(arity_check(f0, a=3, c=5, d=6, discount=1),
                         '"%s" takes 2 parameters. 3 given.' % fname)

    def testArity1(self):
        self.assertEqual(checkarity(f1, (3,), {}), None)
        self.assertEqual(checkarity(f1, (3, 4), {}), None)
        self.assertEqual(checkarity(f1, (), {}),
                         '"f1" takes at least 1 parameters. 0 given.')
        self.assertEqual(checkarity(f1, (4, 5, 6), {}),
                         '"f1" takes at most 2 parameters. 3 given.')
        self.assertEqual(checkarity(f1, (), {'a': 3, 'b': 5}), None)
        self.assertEqual(checkarity(f1, (), {'a': 3, 'c': 5}),
                         '"f1" does not accept "c" parameter.')
        self.assertEqual(checkarity(f1, (), {'a': 3, 'c': 5, 'd': 6}),
                         '"f1" takes at most 2 parameters. 3 given.')

    def testArity2(self):
        self.assertEqual(checkarity(f2, (3,), {}), None)
        self.assertEqual(checkarity(f2, (3,), {'c': 4}), None)
        self.assertEqual(checkarity(f2, (3, 4), {}),
                         '"f2" takes 1 positional parameters. 2 given.')
        self.assertEqual(checkarity(f2, (), {}),
                         '"f2" takes at least 1 parameters. 0 given.')
        self.assertEqual(checkarity(f2, (4, 5, 6), {}),
                         '"f2" takes 1 positional parameters. 3 given.')
        self.assertEqual(checkarity(f2, (), {'a': 3, 'b': 5}), None)
        self.assertEqual(checkarity(f2, (), {'a': 3, 'c': 5}), None)
        self.assertEqual(checkarity(f2, (), {'b': 3, 'c': 5}),
                         '"f2" has missing "a" parameter.')
        self.assertEqual(checkarity(f2, (), {'a': 3, 'c': 5, 'd': 6}), None)


class TestSystemInfo(unittest.TestCase):

    def testMe(self):
        worker = get_actor()
        info = system.process_info(worker.pid)
        system.process_info()
        self.assertTrue(isinstance(info, dict))


# sequential decorator, just for coverage.
@sequential
class TestFunctions(unittest.TestCase):

    def test_convert_bytes(self):
        from pulsar.utils.system import convert_bytes
        self.assertEqual(convert_bytes(None), '#NA')
        self.assertEqual(convert_bytes(4), '4B')
        self.assertEqual(convert_bytes(1024),    '1.0KB')
        self.assertEqual(convert_bytes(1024**2), '1.0MB')
        self.assertEqual(convert_bytes(1024**3), '1.0GB')
        self.assertEqual(convert_bytes(1024**4), '1.0TB')
        self.assertEqual(convert_bytes(1024**5), '1.0PB')
        self.assertEqual(convert_bytes(1024**6), '1.0EB')
        self.assertEqual(convert_bytes(1024**7), '1.0ZB')
        self.assertEqual(convert_bytes(1024**8), '1.0YB')

    def test_nice_number(self):
        self.assertEqual(nice_number(0), 'zero')
        self.assertEqual(nice_number(1), 'one')
        self.assertEqual(nice_number(2), 'two')
        self.assertEqual(nice_number(1, 'bla'), 'one bla')
        self.assertEqual(nice_number(10, 'bla'), 'ten blas')
        self.assertEqual(nice_number(23, 'bla', 'blax'), 'twenty three blax')

    def test_nice_number_large(self):
        self.assertEqual(nice_number(100), 'one hundred')
        self.assertEqual(nice_number(203), 'two hundred and three')
        self.assertEqual(nice_number(4210),
                         'four thousand, two hundred and ten')
        self.assertEqual(nice_number(51345618),
                         'fifty one million, three hundred forty five '
                         'thousand, six hundred and eighteen')

    def test_py_file(self):
        self.assertEqual(py_file('bla.py'), 'bla')
        self.assertEqual(py_file('bla.pyc'), 'bla')
        self.assertEqual(py_file('bla.pyd'), 'bla')
        self.assertEqual(py_file('bla.pyo'), 'bla')
        self.assertEqual(py_file('bla.exe'), None)
        self.assertEqual(py_file('bla.foo.py'), None)
        self.assertEqual(py_file('bla'), 'bla')
        self.assertEqual(py_file('.bla'), None)
        self.assertEqual(py_file('.'), None)

    def __test_import_modules(self):
        self.assertEqual(import_modules(['gggggggggggg']), [])
        mods = import_modules(['pulsar.async.*'])
        self.assertTrue(mods)

    def test_date2timestamp(self):
        d1 = date.today()
        v1 = date2timestamp(d1)
        d2 = datetime.now()
        v2 = date2timestamp(d2)
        self.assertTrue(v2 > v1)
        self.assertEqual(int(v1), v1)



'''Config and Setting classes'''
import os
import pickle
import tempfile
import unittest

import pulsar
from pulsar import get_actor, Config, validate_callable


def connection_made(conn):
    return conn


def post_fork(actor):
    return actor


class TestConfig(unittest.TestCase):

    def testFunction(self):
        cfg = Config()
        worker = get_actor()
        self.assertTrue(cfg.post_fork)
        self.assertEqual(cfg.post_fork(worker), None)
        cfg.set('post_fork', post_fork)
        self.assertEqual(cfg.post_fork(worker), worker)
        cfg1 = pickle.loads(pickle.dumps(cfg))
        self.assertEqual(cfg1.post_fork(worker), worker)

    def __testFunctionFromConfigFile(self):
        # TODO, fails in pypy for some odd reasons
        worker = get_actor()
        cfg = Config()
        self.assertEqual(cfg.connection_made(worker), None)
        self.assertTrue(cfg.import_from_module(__file__))
        self.assertEqual(cfg.connection_made(worker), worker)
        cfg1 = pickle.loads(pickle.dumps(cfg))
        self.assertEqual(cfg1.connection_made(worker), worker)

    def testBadConfig(self):
        cfg = Config()
        self.assertEqual(cfg.config, 'config.py')
        self.assertEqual(cfg.import_from_module('foo/bla/cnkjnckjcn.py'), [])
        cfg.set('config', None)
        self.assertEqual(cfg.config, None)

    def test_exclude(self):
        cfg = Config(exclude=['config'])
        self.assertEqual(cfg.config, 'config.py')
        self.assertEqual(cfg.params['config'], 'config.py')
        self.assertFalse('config' in cfg.settings)

    def testDefaults(self):
        from pulsar.utils import config
        self.assertFalse(config.pass_through(None))
        cfg = Config()
        self.assertEqual(list(sorted(cfg)), list(sorted(cfg.settings)))

        def _():
            cfg.debug = 3
        self.assertRaises(AttributeError, _)
        #
        name = tempfile.mktemp()
        with open(name, 'w') as f:
            f.write('a')
        self.assertRaises(RuntimeError, cfg.import_from_module, name)
        os.remove(name)
        #
        name = '%s.py' % name
        with open(name, 'w') as f:
            f.write('a')
        self.assertRaises(RuntimeError, cfg.import_from_module, name)
        os.remove(name)

    def testSystem(self):
        from pulsar import system
        cfg = Config()
        self.assertEqual(cfg.uid, system.get_uid())
        self.assertEqual(cfg.gid, system.get_gid())
        self.assertEqual(cfg.proc_name, 'pulsar')
        cfg.set('process_name', 'bla')
        self.assertEqual(cfg.proc_name, 'bla')

    def testValidation(self):
        self.assertEqual(pulsar.validate_list((1, 2)), [1, 2])
        self.assertRaises(TypeError, pulsar.validate_list, 'bla')
        self.assertEqual(pulsar.validate_string(' bla  '), 'bla')
        self.assertEqual(pulsar.validate_string(None), None)
        self.assertRaises(TypeError, pulsar.validate_string, [])
        self.assertEqual(pulsar.validate_bool(True), True)
        self.assertEqual(pulsar.validate_bool('true '), True)
        self.assertEqual(pulsar.validate_bool(' false'), False)
        self.assertRaises(TypeError, pulsar.validate_bool, [])
        self.assertRaises(ValueError, pulsar.validate_bool, 'foo')
        self.assertRaises(ValueError, pulsar.validate_pos_int, 'foo')
        self.assertRaises(ValueError, pulsar.validate_pos_int, -1)
        self.assertRaises(ValueError, pulsar.validate_pos_float, 'foo')
        self.assertRaises(ValueError, pulsar.validate_pos_float, -0.001)
        self.assertEqual(pulsar.validate_pos_float('0.101'), 0.101)
        self.assertRaises(TypeError, pulsar.validate_dict, 4)

    def test_validate_callable(self):
        self.assertRaises(TypeError, validate_callable(1), None)
        self.assertRaises(TypeError, validate_callable(1), 4)
        self.assertRaises(TypeError, validate_callable(1), object())

        class test1:
            def __call__(self, arg):
                pass

        class test2:
            def __call__(self, arg1, arg2=None):
                pass

        test = test1()
        self.assertEqual(validate_callable(1)(test), test)
        self.assertRaises(TypeError, validate_callable(2), test)
        test = test2()
        self.assertEqual(validate_callable(1)(test), test)
        self.assertEqual(validate_callable(2)(test), test)
        self.assertRaises(TypeError, validate_callable(3), test)

    def test_methods(self):
        cfg = Config()
        self.assertEqual(cfg.get('sdjcbsjkbcd', 'ciao'), 'ciao')
        d = dict(cfg.items())
        self.assertEqual(len(d), len(cfg))
        sett = cfg.get('debug')
        self.assertTrue(str(sett))
        self.assertEqual(cfg.settings['debug'].default, False)
        cfg.set('debug', True, default=True)
        self.assertEqual(cfg.debug, True)
        self.assertEqual(cfg.settings['debug'].default, True)



import pulsar
from pulsar.apps.wsgi import WSGIServer
from pulsar.apps.ds import pulsards_url

from web import TwitterSite
from twitter import Twitter, PublishTweets


class Server(pulsar.MultiApp):
    '''Create two pulsar applications,

    one for serving the web site and one for streaming tweets
    '''
    # set the default data_store to be pulsar
    cfg = {'data_store': pulsards_url()}

    def build(self):
        # the pubsub channel
        channel = '%s_tweets' % self.name
        yield self.new_app(WSGIServer, callable=TwitterSite(channel))
        yield self.new_app(Twitter, callable=PublishTweets(channel))


if __name__ == '__main__':  # pragma nocover
    Server().start()



from . import wait


class GreenWSGI:
    '''Wraps a WSGI application to be executed on a :class:`.GreenPool`
    '''
    def __init__(self, wsgi, pool):
        self.wsgi = wsgi
        self.pool = pool

    def __call__(self, environ, start_response):
        return self.pool.submit(self._green_handler, environ, start_response)

    def _green_handler(self, environ, start_response):
        if environ['wsgi.input']:
            environ['wsgi.input'] = GreenStream(environ['wsgi.input'])
        return wait(self.wsgi(environ, start_response))


class GreenStream:
    __slots__ = ('stream',)

    def __init__(self, stream):
        self.stream = stream

    def __getattr__(self, name):
        value = getattr(self.stream, name)
        if getattr(value, '__self__', None) is self.stream:
            return _green(value)
        return value


class _green:
    __slots__ = ('value',)

    def __init__(self, value):
        self.value = value

    def __getattr__(self, name):
        return getattr(self.value, name)

    def __call__(self, *args, **kwargs):
        return wait(self.value(*args, **kwargs))



import unittest

from pulsar import async, Future, new_event_loop

__benchmark__ = True
DELAY = 0


def async_func(loop, value):
    p = Future(loop=loop)
    loop.call_later(DELAY, p.set_result, value)
    return p


def sub_sub(loop, num):
    a = yield from async_func(loop, num)
    b = yield from async_func(loop, num)
    return a + b


def sub(loop, num):
    a = yield from async_func(loop, num)
    b = yield from async_func(loop, num)
    c = yield from sub_sub(loop, num)
    return a + b + c


def main(loop, num):
    a = yield from async_func(loop, num)
    b = yield from sub(loop, num)
    c = yield from sub(loop, num)
    return a + b + c


class TestCoroutine33(unittest.TestCase):
    __benchmark__ = True
    __number__ = 100

    def setUp(self):
        self.loop = new_event_loop()

    def test_coroutine(self):
        future = async(main(self.loop, 1), loop=self.loop)
        self.loop.run_until_complete(future)
        self.assertEqual(future.result(), 9)

    def getTime(self, dt):
        return dt - 9*DELAY



'''Tests the wsgi middleware in pulsar.apps.wsgi'''
import unittest

import pulsar
from pulsar.apps.wsgi import Router, RouterParam, route, test_wsgi_environ

from examples.httpbin.manage import HttpBin


class TRouter(Router):
    random = RouterParam(6)


class Router2(Router):
    random1 = RouterParam(5)

    def get(self, request):
        return request.response


class HttpBin2(HttpBin):

    def gzip(self):
        pass    # switch off gzip handler, it is not a route anymore

    @route('get2')
    def get_get(self, request):    # override the get_get handler
        raise pulsar.Http404

    @route()
    def async(self, request):
        future = pulsar.Future()
        future._loop.call_later(0.5, lambda: future.set_result(['Hello!']))
        return future

    @route()
    def post_async(self, request):
        future = pulsar.Future()
        future._loop.call_later(0.5, lambda: future.set_result(['Hello!']))
        return future


class MyRouter(Router):

    @route('/<id>')
    def get_elem(self, request):
        return request.response

    @route('/<id>')
    def post_elem(self, request):
        return request.response


class HttpBin3(HttpBin):

    @route('new', position=0)
    def new(self, request):
        return self.info_data_response(request)

    @route(method='post', title='Returns POST data', position=-1)
    def post_post(self, request):
        return self.info_data_response(request)


class TestRouter(unittest.TestCase):

    def router(self, path='/'):
        class testRouter(Router):
            response_content_types = RouterParam(('text/html',
                                                  'text/plain',
                                                  'application/json'))

            def get(self, request):
                return 'Hello World!'

            @route()
            def bla(self, request):
                return 'This is /bla route'

            @route('/foo')
            def xxx(self, request):
                return 'This is /foo route'

            @route()
            def post_pluto(self, request):
                return 'This is /pluto POST route'

        router = testRouter(path)
        self.assertEqual(len(router.routes), 3)
        return router

    def test_router(self):
        router = self.router()
        self.assertEqual(router.route.path, '/')
        handler, urlargs = router.resolve('')
        self.assertEqual(handler, router)
        self.assertEqual(urlargs, {})
        #
        handler, urlargs = router.resolve('bla')
        self.assertNotEqual(handler, router)
        self.assertEqual(urlargs, {})

    def test_derived(self):
        self.assertTrue('gzip' in HttpBin.rule_methods)
        self.assertFalse('gzip' in HttpBin2.rule_methods)

    def test_override(self):
        self.assertTrue('get_get' in HttpBin.rule_methods)
        self.assertEqual(HttpBin.rule_methods['get_get'][0].rule, 'get')
        self.assertTrue('get_get' in HttpBin2.rule_methods)
        self.assertEqual(HttpBin2.rule_methods['get_get'][0].rule, 'get2')
        # The position in the ordered dict should be the same too
        all = list(HttpBin.rule_methods)
        all2 = list(HttpBin2.rule_methods)
        self.assertEqual(all2.index('get_get'), all.index('get_get'))

    def test_override_change_position(self):
        self.assertTrue('post_post' in HttpBin.rule_methods)
        self.assertEqual(HttpBin.rule_methods['post_post'][0].rule, 'post')
        self.assertTrue('post_post' in HttpBin3.rule_methods)
        self.assertEqual(HttpBin3.rule_methods['post_post'][0].rule, 'post')
        # The position in the ordered dict should be the same too
        all = list(HttpBin.rule_methods)
        all3 = list(HttpBin3.rule_methods)
        self.assertEqual(all3.index('new'), 1)
        self.assertTrue(all3.index('post_post') < all.index('post_post'))

    def test_path_method(self):
        router = Router('/root',
                        Router('a', get=lambda r: ['route a']))
        self.assertEqual(router.path(), '/root')
        self.assertEqual(router.route.is_leaf, True)
        child, args = router.resolve('root/a')
        self.assertFalse(args)
        self.assertEqual(child.parent, router)
        self.assertEqual(child.path(), '/root/a')

    def test_router_count(self):
        self.assertTrue(HttpBin2.rule_methods)
        async = HttpBin2.rule_methods.get('async')
        self.assertTrue(async)
        self.assertEqual(async.method, 'get')
        self.assertEqual(str(async.rule), '/async')
        async = HttpBin2.rule_methods.get('post_async')
        self.assertTrue(async)
        self.assertEqual(async.method, 'post')
        self.assertEqual(str(async.rule), '/async')
        #
        router = HttpBin2('/')
        self.assertEqual(router.name, '')
        router = HttpBin2('/', name='root')
        self.assertEqual(router.name, 'root')
        async = router.get_route('async')
        self.assertTrue(async)
        # It has both get and post methods
        self.assertTrue(async.get)
        self.assertTrue(async.post)

    def test_router_child(self):
        router = TRouter('/', HttpBin2('bin'), random=9)
        self.assertEqual(router.name, '')
        self.assertEqual(len(router.routes), 1)
        self.assertEqual(router.random, 9)
        self.assertEqual(router.root, router)
        child = router.get_route('binx')
        self.assertFalse(child)
        child = router.get_route('bin')
        self.assertTrue(child)
        self.assertEqual(router.root, child.root)
        self.assertTrue(router.has_parent(router))
        self.assertFalse(router.has_parent(child))
        self.assertTrue(child.has_parent(router))

    def test_child_methods(self):
        router = TRouter('/', HttpBin2('bin'), name='home')
        self.assertTrue('random' in router.defaults)
        #
        self.assertEqual(router.name, 'home')
        child = router.get_route('bin')
        self.assertTrue(child)
        #
        async = router.get_route('async')
        self.assertFalse('random' in async.defaults)
        self.assertTrue(async)
        self.assertEqual(async.root, router)
        self.assertEqual(async.parent, child)
        self.assertEqual(async.random, 6)
        #
        # It has both get and post methods
        self.assertTrue(async.get)
        self.assertTrue(async.post)

    def test_rule(self):
        router = TRouter('/', HttpBin2('bin'))
        self.assertEqual(repr(router), '/')
        self.assertEqual(router.rule, router.full_route.rule)

    def test_remove_child(self):
        router = TRouter('/', HttpBin2('bin'))
        child = router.routes[0]
        self.assertEqual(child.path(), '/bin')
        self.assertEqual(child.parent, router)
        router.remove_child(child)
        self.assertFalse(router.routes)
        self.assertEqual(child.parent, None)

    def test_default(self):
        router = Router2('/', TRouter('foo'))
        self.assertTrue(router.get)
        foo = router.get_route('foo')
        self.assertRaises(AttributeError, lambda: foo.get)
        self.assertEqual(foo.random1, 5)
        router = Router2('/', TRouter('foo'), random1=10)
        foo = router.get_route('foo')
        self.assertRaises(AttributeError, lambda: foo.get)
        self.assertEqual(foo.random1, 10)

    def test_multiple_methods(self):
        router = MyRouter('/')
        child = router.get_route('elem')
        self.assertTrue(child)
        self.assertTrue(child.get.__name__, 'get_elem')
        self.assertTrue(child.post.__name__, 'post_elem')

    def test_response_wrapper(self):

        def response_wrapper(callable, response):
            raise pulsar.PermissionDenied('Test Response Wrapper')

        router = HttpBin('/', response_wrapper=response_wrapper)

        environ = test_wsgi_environ()
        try:
            router(environ, None)
        except pulsar.PermissionDenied as exc:
            self.assertEqual(str(exc), 'Test Response Wrapper')
        else:
            raise RuntimeError

        environ = test_wsgi_environ('/get')
        try:
            router(environ, None)
        except pulsar.PermissionDenied as exc:
            self.assertEqual(str(exc), 'Test Response Wrapper')
        else:
            raise RuntimeError



import os
import unittest

from pulsar.utils.httpurl import hasextensions
from pulsar.utils import httpurl


class TestPythonHttpParser(unittest.TestCase):

    @classmethod
    def parser(cls, **kwargs):
        return httpurl.HttpParser(**kwargs)

    def __test_amazon_protocol_error(self):
        all = []
        for n in range(18):
            with open(os.path.join(os.path.dirname(__file__), 'data',
                                   'data%d.dat' % n), 'rb') as f:
                all.append(f.read())
        #
        p = self.parser()
        for chunk in all:
            self.assertEqual(p.execute(chunk, len(chunk)), len(chunk))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())

    def test_client_200_OK_no_headers(self):
        p = self.parser()
        data = b'HTTP/1.1 200 OK\r\n\r\n'
        self.assertEqual(p.execute(data, len(data)), len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())
        self.assertFalse(p.is_partial_body())
        self.assertFalse(p.is_message_complete())

    def test_simple_server_message(self):
        p = self.parser()
        self.assertFalse(p.is_headers_complete())
        data = b'GET /forum/bla?page=1#post1 HTTP/1.1\r\n\r\n'
        self.assertEqual(p.execute(data, len(data)), len(data))
        self.assertEqual(p.get_fragment(), 'post1')
        self.assertTrue(p.is_message_begin())
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_complete())
        self.assertFalse(p.get_headers())

    def test_bad_combinations(self):
        p = self.parser()
        # body with no headers not valid
        data = b'GET /get HTTP/1.1\r\nciao'
        self.assertEqual(p.execute(data, len(data)), len(data))
        self.assertFalse(p.is_headers_complete())
        # self.assertTrue(p.is_message_begin())
        self.assertFalse(p.is_message_complete())
        self.assertEqual(p.execute(b'x', 1), 1)

    def test_client_connect(self):
        p = self.parser()
        data = b'HTTP/1.1 200 Connection established\r\n\r\n'
        self.assertEqual(p.execute(data, len(data)), len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())
        self.assertFalse(p.is_partial_body())
        self.assertFalse(p.is_message_complete())

    def test_client_no_body(self):
        p = self.parser()
        data = (b'HTTP/1.1 200 OK\r\n'
                b'Connection: keep-alive\r\n\r\n')
        self.assertEqual(p.execute(data, len(data)), len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())
        self.assertFalse(p.is_partial_body())
        self.assertFalse(p.is_message_complete())

    def testBadFirstLine(self):
        p = self.parser()
        self.assertFalse(p.is_headers_complete())
        data = b'GET HTTP/1.1\r\n\r\n'
        self.assertNotEqual(p.execute(data, len(data)), len(data))
        p = self.parser()
        data = b'get /foo HTTP/1.1\r\n\r\n'
        self.assertNotEqual(p.execute(data, len(data)), len(data))
        p = self.parser()
        data = b'GET /foo FTP/1.1\r\n\r\n'
        self.assertNotEqual(p.execute(data, len(data)), len(data))
        p = self.parser()
        data = b'GET get/ HTTP/1-x\r\n'
        self.assertNotEqual(p.execute(data, len(data)), len(data))

    def testPartialFirstLine(self):
        p = self.parser()
        data = b'GET /get H'
        self.assertEqual(p.execute(data, len(data)), len(data))
        self.assertFalse(p.is_headers_complete())
        data = b'TTP/1.1\r\n\r\n'
        p.execute(data, len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_complete())
        self.assertFalse(p.get_headers())

    def testBadHeader(self):
        p = self.parser()
        data = b'GET /get HTTP/1.1\r\nbla\0: bar\r\n\r\n'
        self.assertNotEqual(p.execute(data, len(data)), len(data))
        #
        p = self.parser()
        data = b'GET /test HTTP/1.1\r\nfoo\r\n\r\n'
        self.assertEqual(p.execute(data, len(data)), len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())
        self.assertTrue(p.is_message_complete())
        headers = p.get_headers()
        self.assertEqual(len(headers), 0)

    def testHeaderOnly(self):
        p = self.parser()
        data = b'GET /test HTTP/1.1\r\nHost: 0.0.0.0=5000\r\n\r\n'
        p.execute(data, len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())
        self.assertTrue(p.is_message_complete())

    def testContentLength(self):
        p = self.parser()
        data = b'GET /test HTTP/1.1\r\n'
        p.execute(data, len(data))
        self.assertFalse(p.is_headers_complete())
        data = b'Content-Length: 4\r\n\r\n'
        p.execute(data, len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())
        self.assertFalse(p.is_message_complete())
        data = b'cia'
        p.execute(data, len(data))
        self.assertFalse(p.is_message_complete())
        data = b'o'
        p.execute(data, len(data))
        self.assertTrue(p.is_message_complete())

    def test_ouble_header(self):
        p = self.parser()
        data = b'GET /test HTTP/1.1\r\n'
        p.execute(data, len(data))
        data = b'Accept: */*\r\n'
        p.execute(data, len(data))
        data = b'Accept: jpeg\r\n\r\n'
        p.execute(data, len(data))
        self.assertTrue(p.is_headers_complete())
        self.assertTrue(p.is_message_begin())
        self.assertTrue(p.is_message_complete())
        headers = p.get_headers()
        self.assertEqual(len(headers), 1)
        self.assertEqual(headers.get('Accept'), ['*/*', 'jpeg'])

    def test_connect(self):
        p = self.parser()
        data = b'HTTP/1.1 200 Connection established\r\n\r\n'
        self.assertEqual(p.execute(data, len(data)), len(data))


@unittest.skipUnless(hasextensions, 'Requires C extensions')
class TestCHttpParser(TestPythonHttpParser):

    @classmethod
    def parser(cls, **kwargs):
        return httpurl.CHttpParser(**kwargs)



'''Tests git info.'''
import unittest

from pulsar.utils.version import gitrepo


class TestGit(unittest.TestCase):

    def test_pulsar(self):
        info = gitrepo()
        self.assertTrue(info)
        self.assertTrue(info['branch'])
        self.assertIsInstance(info['head'], dict)
        self.assertIsInstance(info['remotes'], list)
        remote = info['remotes'][0]
        self.assertIsInstance(remote, dict)
        self.assertEqual(remote['name'], 'origin')



import os
from datetime import date

from agile.release import ReleaseManager
version_file = os.path.join(os.path.dirname(__file__), 'pulsar', '__init__.py')

note_file = 'releases/notes.rst'


def write_notes(manager, release):
    filename = 'CHANGELOG.rst'
    dt = date.today()
    dt = dt.strftime('%Y-%b-%d')
    body = ['Ver. %s - %s' % (release['tag_name'], dt),
            '='*28,
            release['body']]

    if os.path.isfile(filename):
        with open(filename, 'r') as file:
            body.append('\n')
            body.append(file.read())

    with open(filename, 'w') as file:
        file.write('\n'.join(body))

    manager.logger.info('Added notes to changelog')


if __name__ == '__main__':
    ReleaseManager(config='release.py').start()



from django.conf.urls import include, url
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns

from .views import home

urlpatterns = [
    url(r'^$', home),
    url(r'^admin/', include(admin.site.urls)),
]

urlpatterns.extend(staticfiles_urlpatterns())



'''Tests actor and actor proxies.'''
import unittest
from time import time

from functools import partial

import pulsar
from pulsar import send, spawn, async_while, TcpServer, Connection
from pulsar.apps.test import ActorTestMixin, dont_run_with_thread

from examples.echo.manage import EchoServerProtocol


def add(actor, a, b):
    return (actor.name, a+b)


def spawn_actor_from_actor(actor, name):
    actor2 = yield from spawn(name=name)
    pong = yield from send(actor2, 'ping')
    assert pong == 'pong', 'no pong from actor'
    t1 = time()
    # cover the notify from a fron actor
    t2 = yield from send(actor2, 'notify', {})
    assert t2 >= t1

    return actor2.aid


class create_echo_server:
    '''partial is not picklable in python 2.6'''
    def __init__(self, address):
        self.address = address

    def __call__(self, actor):
        '''Starts an echo server on a newly spawn actor'''
        server = TcpServer(partial(Connection, EchoServerProtocol),
                           actor._loop, self.address)
        yield from server.start_serving()
        actor.servers['echo'] = server
        actor.extra['echo-address'] = server.address
        actor.bind_event('stopping', self._stop_server)
        return actor

    def _stop_server(self, actor):
        yield from actor.servers['echo'].close()
        return actor


class TestActorThread(ActorTestMixin, unittest.TestCase):
    concurrency = 'thread'

    def test_spawn_and_interact(self):
        name = 'pluto-%s' % self.concurrency
        proxy = yield from self.spawn_actor(name=name)
        self.assertEqual(proxy.name, name)
        yield from self.async.assertEqual(send(proxy, 'ping'), 'pong')
        yield from self.async.assertEqual(send(proxy, 'echo', 'Hello!'),
                                          'Hello!')
        n, result = yield from send(proxy, 'run', add, 1, 3)
        self.assertEqual(n, name)
        self.assertEqual(result, 4)

    def test_info(self):
        name = 'pippo-%s' % self.concurrency
        proxy = yield from self.spawn_actor(name=name)
        self.assertEqual(proxy.name, name)
        info = yield from send(proxy, 'info')
        self.assertTrue('actor' in info)
        ainfo = info['actor']
        self.assertEqual(ainfo['is_process'], self.concurrency == 'process')

    def test_simple_spawn(self):
        '''Test start and stop for a standard actor on the arbiter domain.'''
        proxy = yield from self.spawn_actor(
            name='simple-actor-on-%s' % self.concurrency)
        arbiter = pulsar.get_actor()
        proxy_monitor = arbiter.get_actor(proxy.aid)
        self.assertEqual(proxy_monitor, proxy)
        yield from self.async.assertEqual(send(proxy, 'ping'), 'pong')
        yield from self.async.assertEqual(send(proxy.proxy, 'echo', 'Hello!'),
                                          'Hello!')
        # We call the ActorTestMixin.stop_actors method here, since the
        # ActorTestMixin.tearDown method is invoked on the test-worker domain
        # (here we are in the arbiter domain)
        yield from self.stop_actors(proxy)
        is_alive = yield from async_while(3, proxy_monitor.is_alive)
        self.assertFalse(is_alive)

    def test_spawn_from_actor(self):
        proxy = yield from self.spawn_actor(
            name='spawning-actor-%s' % self.concurrency)
        arbiter = pulsar.get_actor()
        self.assertTrue(repr(proxy).startswith('spawning-actor-'))
        self.assertEqual(proxy, proxy.proxy)
        proxy_monitor = arbiter.get_actor(proxy.aid)
        self.assertFalse(proxy != proxy_monitor)
        #
        # do the spawning
        name = 'spawned-actor-%s-from-actor' % self.concurrency
        aid = yield from send(proxy, 'run', spawn_actor_from_actor, name)
        self.assertTrue(aid)
        proxy_monitor2 = arbiter.get_actor(aid)
        self.assertEqual(proxy_monitor2.name, name)
        self.assertNotEquals(proxy_monitor, proxy_monitor2)
        #
        # stop them
        yield from self.stop_actors(proxy, proxy_monitor2)
        is_alive = yield from async_while(3, proxy_monitor.is_alive)
        self.assertFalse(is_alive)
        is_alive = yield from async_while(3, proxy_monitor2.is_alive)
        self.assertFalse(is_alive)

    def test_config_command(self):
        proxy = yield from self.spawn_actor(
            name='actor-test-config-%s' % self.concurrency)
        arbiter = pulsar.get_actor()
        proxy_monitor = arbiter.get_actor(proxy.aid)
        result = yield from send(proxy, 'config', 'khjkh', 'name')
        self.assertEqual(result, None)
        result = yield from send(proxy, 'config', 'get', 'concurrency')
        self.assertEqual(result, self.concurrency)
        result = yield from send(proxy, 'config', 'get', 'concurrency', 'foo')
        self.assertEqual(result, None)
        #
        result = yield from send(proxy, 'config', 'set', 'max_requests',
                                 1000, 1000)
        self.assertEqual(result, None)
        result = yield from send(proxy, 'config', 'set', 'max_requests',
                                 1000)
        self.assertEqual(result, True)
        result = yield from send(proxy, 'config', 'get', 'max_requests')
        self.assertEqual(result, 1000)
        #
        yield from self.stop_actors(proxy)
        is_alive = yield from async_while(3, proxy_monitor.is_alive)
        self.assertFalse(is_alive)


@dont_run_with_thread
class TestActorProcess(TestActorThread):
    concurrency = 'process'
